import os
import subprocess
import tempfile
from dataclasses import dataclass
from importlib.resources import path as res_path
from typing import Iterable, List, Union

import numpy as np
import pandas as pd
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.dataset import write_features
from ertk.preprocessing.base import FeatureExtractor


@dataclass
class OpenXBOWExtractorConfig(ERTKConfig):
    xbowargs: List[str] = MISSING


class OpenXBOWExtractor(
    FeatureExtractor, fname="openxbow", config=OpenXBOWExtractorConfig
):
    config: OpenXBOWExtractorConfig

    def __init__(self, config: OpenXBOWExtractorConfig) -> None:
        super().__init__(config)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        raise NotImplementedError()

    def process_batch(
        self,
        batch: Union[Iterable[np.ndarray], np.ndarray],
        **kwargs,
    ) -> List[np.ndarray]:

        _, tmpin = tempfile.mkstemp(prefix="openxbow_", suffix=".csv")
        _, tmpout = tempfile.mkstemp(prefix="openxbow_", suffix=".csv")

        xs = list(batch)
        names = [str(i) for i in range(len(xs))]
        write_features(tmpin, xs, names=names)

        with res_path("ertk.preprocessing.openxbow", "openXBOW.jar") as jar:
            xbow_args = [
                "java",
                "-jar",
                str(jar),
                "-i",
                tmpin,
                "-o",
                tmpout,
                "-attributes",
                f"n1[{xs[0].shape[1]}]",
                "-csvSep",
                ",",
                "-writeName",
                "-noLabels",
                *self.config.xbowargs,
            ]

        if subprocess.call(xbow_args) != 0:
            raise RuntimeError("openXBOW returned an error")
        os.remove(tmpin)
        data: pd.DataFrame = pd.read_csv(
            tmpout, header=None, quotechar="'", dtype={0: str}
        )
        os.remove(tmpout)

        return list(data.iloc[:, 1:].to_numpy())

    @property
    def dim(self) -> int:
        return super().dim

    @property
    def is_sequence(self) -> bool:
        return False



from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Type, TypeVar, Union, cast

import yaml
from omegaconf import DictConfig, OmegaConf

from ertk.utils import PathOrStr

T = TypeVar("T", bound="ERTKConfig")


@dataclass
class ERTKConfig(ABC):
    def to_dictconfig(self) -> DictConfig:
        return DictConfig(self)

    @classmethod
    def from_config(cls: Type[T], config: Any) -> T:
        schema = OmegaConf.structured(cls)
        return cast(T, OmegaConf.merge(schema, config))

    @classmethod
    def from_yaml_file(cls: Type[T], path: PathOrStr) -> T:
        config = OmegaConf.load(Path(path))
        return cls.from_config(config)


def get_arg_mapping_multi(s: str) -> Dict[str, List[Any]]:
    """Given a string mapping from the command-line, returns a dict
    representing that mapping.

    The string form of the mapping is::
        key:value[,key:value]+

    Duplicate keys will be mapped to a list of values.

    Parameters
    ----------
    s: str
        String representing the mapping. It cannot contain spaces or
        shell symbols (unless escaped).

    Returns
    -------
    mapping: dict
        A dictionary mapping keys to lists of values from the string.
    """
    mapping: Dict[str, List[str]] = {}
    for cls in s.split(","):
        key, val = cls.split(":")
        if key in mapping:
            mapping[key].append(val)
        else:
            mapping[key] = [val]
    return mapping


def get_arg_mapping(s: Union[Path, str]) -> Dict[str, Any]:
    """Given a mapping on the command-line, returns a dict representing
    that mapping. Mapping can be a string or a more complex YAML file.

    The string form of the mapping is:
        key:value[,key:value]+

    Parameters
    ----------
    s: PathLike or str
        String representing the mapping or path to YAML containing
        mapping. If a string, it cannot contain spaces or shell symbols
        (unless escaped).

    Returns
    -------
    mapping: dict
        A dictionary mapping keys to values from the string.
    """
    if isinstance(s, Path) or Path(s).exists():
        with open(s) as fid:
            return yaml.safe_load(fid) or {}
    return {k: v[0] if len(v) == 1 else v for k, v in get_arg_mapping_multi(s).items()}



"""TensorFlow implementations of various models."""

import importlib
from functools import partial

from tensorflow.keras.models import Model

from ertk.tensorflow.utils import TFModelFunction


def get_tf_model_fn(name: str, **kwargs) -> TFModelFunction:
    """Get a TensorFlow model creation function.

    Parameters
    ----------
    name: str
        The model name. This should be the name of the module that
        implements the model.
    **kwargs:
        Keyword arguments to provide to the model creation function.

    Returns
    -------
    model_fn: callable
        A method that takes arguments and returns a Model instance.
    """
    try:
        module = importlib.import_module(f".{name}", __package__)
        model_fn = getattr(module, "model")
    except (ImportError, AttributeError) as e:
        raise ValueError(
            f"{name} does not correspond to a valid TensorFlow model."
        ) from e
    return partial(model_fn, **kwargs)


def get_tf_model(name: str, **kwargs) -> Model:
    """Get a TensorFlow model by name.

    Parameters
    ----------
    name: str
        The model name. This should be the name of the module that
        implements the model.
    **kwargs:
        Keyword arguments to provide to the model creation function.

    Returns
    -------
    model: Model
        The uncompiled TensorFlow model.
    """
    return get_tf_model_fn(name, **kwargs)()



"""Implementation of short model from [1].

[1] M. G. de Pinto, M. Polignano, P. Lops, and G. Semeraro, 'Emotions
Understanding Model from Spoken Language using Deep Neural Networks and
Mel-Frequency Cepstral Coefficients', in 2020 IEEE Conference on
Evolving and Adaptive Intelligent Systems (EAIS), May 2020, pp. 1–5,
doi: 10.1109/EAIS48028.2020.9122698.
"""


from tensorflow.keras import Model
from tensorflow.keras.layers import (
    Conv1D,
    Dense,
    Dropout,
    Flatten,
    Input,
    MaxPool1D,
    Reshape,
)


def model(n_features: int, n_classes: int) -> Model:
    inputs = Input((n_features,))
    x = Reshape((n_features, 1))(inputs)
    # x = Conv1D(128, 40, activation="relu", padding="same")(x)
    # x = Dropout(0.2)(x)
    # x = MaxPool1D(8)(x)
    x = Conv1D(64, 5, activation="relu", padding="same")(x)
    x = Dropout(0.2)(x)
    x = Flatten()(x)
    x = Dense(n_classes, activation="softmax")(x)
    return Model(inputs=inputs, outputs=x)



import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from sklearn.base import ClassifierMixin
from sklearn.model_selection import BaseCrossValidator, cross_validate
from sklearn.pipeline import Pipeline

from ertk.train import get_pipeline_params, get_scores
from ertk.utils import ScoreFunction, filter_kwargs

_logger = logging.getLogger("ertk.sklearn.classification")


def sk_cross_validate(
    clf: ClassifierMixin,
    x: np.ndarray,
    y: Optional[np.ndarray] = None,
    *,
    groups: Optional[np.ndarray] = None,
    cv: BaseCrossValidator = None,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    verbose: int = 0,
    n_jobs: int = 1,
    fit_params: Dict[str, Any] = {},
):
    if isinstance(clf, Pipeline):
        fit_params = get_pipeline_params(fit_params, clf)
    else:
        fit_params = filter_kwargs(fit_params, clf.fit)
    _logger.debug(f"cross_validate(): filtered fit_params={fit_params}")

    return cross_validate(
        clf,
        x,
        y,
        cv=cv,
        scoring=scoring,
        groups=groups,
        n_jobs=n_jobs,
        verbose=verbose,
        fit_params=fit_params,
        error_score="raise",
    )


def sk_train_val_test(
    clf: ClassifierMixin,
    train_data: Tuple[np.ndarray, ...],
    valid_data: Tuple[np.ndarray, ...],
    test_data: Optional[Tuple[np.ndarray, ...]] = None,
    verbose: int = 0,
    scoring: Union[
        str, List[str], Dict[str, ScoreFunction], Callable[..., float]
    ] = "accuracy",
    fit_params: Dict[str, Any] = {},
):
    if isinstance(clf, Pipeline):
        fit_params = get_pipeline_params(fit_params, clf)
    else:
        fit_params = filter_kwargs(fit_params, clf.fit)
    _logger.debug(f"train_val_test(): filtered fit_params={fit_params}")

    if test_data is None:
        test_data = valid_data

    if verbose > 0:
        print(f"Fitting classifier {clf}")
    clf = clf.fit(train_data[0], train_data[1], **fit_params)
    if verbose > 0:
        print(f"Getting predictions from classifier {clf}")
    y_pred = clf.predict(test_data[0])
    scores = get_scores(scoring, y_pred, test_data[1])
    scores = {f"test_{k}": v for k, v in scores.items()}
    return scores



from typing import List, Optional, Sequence, Tuple, Union, overload

import numpy as np


def make_array_array(x: List[np.ndarray]) -> np.ndarray:
    """Helper function to make an array of arrays. The returned array
    has `shape==(len(x),)` and `dtype==object`.
    """
    arr = np.empty(len(x), dtype=object)
    for i, a in enumerate(x):
        arr[i] = a
    return arr


def check_3d(arrays: Union[Sequence[np.ndarray], np.ndarray]) -> None:
    """Checks if an array is 3D or each array in a list is 2D. Raises an
    exception if this isn't the case.
    """
    if any(len(x.shape) != 2 for x in arrays):
        raise ValueError("arrays must be 3D (contiguous or vlen).")


def is_mono_audio(x: np.ndarray) -> bool:
    """Checks if an array represents mono audio data. The conditions for
    this are that x can be squeezed to 1D.

    Note that this method will also return True for '2D' sequences of
    length 1, so care should be taken for edge cases.
    """
    return np.squeeze(x).ndim == 1


def frame_array(
    x: np.ndarray,
    frame_size: int,
    frame_shift: int,
    pad: bool = False,
    axis: int = 0,
    copy: bool = True,
) -> np.ndarray:
    """Frames an array over a given axis with optional padding and
    copying.

    Parameters
    ----------
    x: np.ndarray
        Array to frame.
    frame_size: int
        Width of frame.
    frame_shift: int
        Amount each window is moved for subsequent frames.
    pad: bool
        If the frames don't neatly fit the axis length, this indicated
        whether to pad or cut the remaining length. Default is `False`
        which means any remaining length is cut. If `pad=True` then any
        remaining length is padded with zeros and an additional frame is
        produced. If frames neatly fit into the axis length then this
        option has no effect.
    axis: int
        Axis to frame over. Default is 0.
    copy: bool
        Whether to copy and return a contiguous array (more memory
        usage). Default is `True` so that a contiguous copy is retured.
        Note: if `copy=False` it is assumed that the input array is
        contiguous so that a view with modified strides can be properly
        created.

    Returns
    -------
    frames: np.ndarray
        The resulting frames. If `copy=False` then these are read-only
        views into x, otherwise a contiguous array is returned.
    """
    idx_slice = tuple(slice(None) for _ in range(axis))

    num_frames = (x.shape[axis] - frame_size) // frame_shift + 1
    if num_frames <= 0:
        if not pad:
            raise ValueError(
                "The length of the sequence is shorter than frame_size, but pad=False, "
                "so no frames will be generated."
            )
        num_frames = 0
    remainder = (x.shape[axis] - frame_size) % frame_shift
    if remainder != 0:
        num_frames += 1 if pad else 0
    # Insert new dim before axis with num_frames, and axis is replaced
    # with the size of each frame.
    new_shape = x.shape[:axis] + (num_frames, frame_size) + x.shape[axis + 1 :]

    if copy:
        out = np.zeros(new_shape, dtype=x.dtype)
        for i in range(0, x.shape[axis] - frame_size + 1, frame_shift):
            out[idx_slice + (i // frame_shift,)] = x[
                idx_slice + (slice(i, i + frame_size),)
            ]
        if remainder != 0 and pad:
            if num_frames == 1:
                left = x.shape[axis]
            else:
                left = frame_size + remainder - frame_shift
            # out[..., -1, 0:left, ...] = x[..., -left:, ...]
            out[idx_slice + (-1, slice(0, left))] = x[idx_slice + (slice(-left, None),)]

        return out

    from numpy.lib.stride_tricks import as_strided

    if remainder != 0:
        if pad:
            # Padding creates a new array copy
            raise ValueError("pad must be False when copy=False")
        # x[:, :, ..., 0:N-r, :, ...]
        x = x[idx_slice + (slice(0, x.shape[axis] - remainder),)]
    # Add a new stride for the num_frames part of shape which moves by
    # frame_shift for each frame. The stride within each frame is the
    # same as for the original axis.
    new_strides = x.strides[:axis] + (frame_shift * x.strides[axis],) + x.strides[axis:]
    return as_strided(x, new_shape, new_strides, writeable=False)


def frame_arrays(
    arrays: Union[List[np.ndarray], np.ndarray],
    frame_size: int,
    frame_shift: int,
    max_frames: Optional[int] = None,
    vlen: bool = False,
) -> np.ndarray:
    """Creates sequences of frames from the given arrays. This is mainly
    a convenience wrapper around `frame_array()` for instance arrays.

    Parameters
    ----------
    arrays: list of np.ndarray
        The arrays to process.
    frame_size: int
        Length of each frame.
    frame_shift: int
        Amount each frame is shifted.
    max_frames: int, optional
        If given, only the first max_frames are returned for each array.
    vlen: bool
        Whether to return a contiguous array if `vlen=False` (default)
        or an array of variable length arrays.

    Returns
    -------
    framed_arrays:
        Framed arrays.
    """
    check_3d(arrays)

    if max_frames is None:
        max_len = max(len(x) for x in arrays)
        max_frames = (max_len - frame_size) // frame_shift + 1

    # Generator
    arrs = (
        frame_array(x, frame_size, frame_shift, axis=0, pad=True, copy=True)[
            :max_frames
        ]
        for x in arrays
    )
    if vlen:
        return make_array_array(list(arrs))

    n_features = arrays[0].shape[1]
    framed_arrays = np.zeros(
        (len(arrays), max_frames, frame_size, n_features), dtype=arrays[0].dtype
    )
    for i, arr in enumerate(arrs):
        framed_arrays[i, : len(arr)] = arr
    return framed_arrays


def pad_array(
    x: np.ndarray, to_multiple: int = None, to_size: int = None, axis: int = 0
):
    """Pads an array either to a multiple of `to_multiple` or to the
    exact length `to_size` along `axis`.


    Parameters
    ----------
    x: np.ndarray
        The array to pad.
    to_multiple: int, optional
        If given, `x` is padded so that it's length is the nearest
        larger multiple of `to_multiple`.
    to_size: int, optional
        If given, `x` is padded so that it's length is exactly
        `to_size`. Default is None, and instead `to_multiple` is used.
        Exactly one of `to_size` and `to_multiple` must be given.
    axis: int
        The axis along which to pad. Default is axis 0.

    Returns
    -------
    padded: np.ndarray
        The padded array.
    """
    if to_multiple is not None and to_size is not None:
        raise ValueError("Only one of `to_multiple` and `to_size` should be given.")

    if to_size is not None:
        to_pad = to_size - x.shape[axis]
        if to_pad < 0:
            raise ValueError("The length of `x` is already greater than `to_size`.")
    elif to_multiple is not None:
        to_pad = int(np.ceil(x.shape[axis] / to_multiple)) * to_multiple - x.shape[axis]
    else:
        raise ValueError("One of `to_multiple` and `to_size` must be given.")

    if to_pad == 0:
        return x.copy()  # A copy is expected when padding

    pad = [(0, 0) for _ in x.shape]
    pad[axis] = (0, to_pad)
    return np.pad(x, pad)


@overload
def pad_arrays(arrays: List[np.ndarray], pad: int) -> List[np.ndarray]:
    pass


@overload
def pad_arrays(arrays: np.ndarray, pad: int) -> np.ndarray:
    pass


def pad_arrays(arrays: Union[List[np.ndarray], np.ndarray], pad: int = 32):
    """Pads each array to the nearest multiple of `pad` greater than the
    array size. Assumes axis 0 of each sub-array, or axis 1 of x, is
    the time axis. This is mainly a wrapper around `pad_array()` for
    instance arrays.
    """
    if isinstance(arrays, np.ndarray) and len(arrays.shape) > 1:
        # Contiguous 2D/3D
        return pad_array(arrays, to_multiple=pad, axis=1)
    new_arrays = [pad_array(x, to_multiple=pad, axis=0) for x in arrays]
    if isinstance(arrays, np.ndarray):
        if all(x.shape == new_arrays[0].shape for x in new_arrays):
            # Contiguous array
            return np.stack(new_arrays)
        # Array of variable-length arrays
        return make_array_array(new_arrays)
    # List
    return new_arrays


@overload
def clip_arrays(arrays: List[np.ndarray], length: int, copy: bool) -> List[np.ndarray]:
    pass


@overload
def clip_arrays(arrays: np.ndarray, length: int, copy: bool) -> np.ndarray:
    pass


def clip_arrays(
    arrays: Union[List[np.ndarray], np.ndarray], length: int, copy: bool = True
):
    """Clips each array to the specified maximum length."""
    if isinstance(arrays, np.ndarray):
        if len(arrays.shape) > 1:
            return arrays[:, :length, ...].copy() if copy else arrays[:, :length, ...]
        new_arrays = [x[:length].copy() if copy else x[:length] for x in arrays]
        if all(x.shape == new_arrays[0].shape for x in new_arrays):
            # Return contiguous array
            return np.stack(new_arrays)
        return make_array_array(new_arrays)
    return [x[:length].copy() if copy else x[:length] for x in arrays]


@overload
def transpose_time(arrays: List[np.ndarray]) -> List[np.ndarray]:
    pass


@overload
def transpose_time(arrays: np.ndarray) -> np.ndarray:
    pass


def transpose_time(arrays: Union[List[np.ndarray], np.ndarray]):
    """Transpose the time and feature axis of each array. Requires each
    array be 2-D.

    Note: This function modifies the arrays in-place.
    """
    check_3d(arrays)
    if isinstance(arrays, np.ndarray) and len(arrays.shape) == 3:
        arrays = arrays.transpose(0, 2, 1)
    else:
        for i in range(len(arrays)):
            arrays[i] = arrays[i].transpose()
    assert all(x.shape[0] == arrays[0].shape[0] for x in arrays)
    return arrays


def shuffle_multiple(
    *arrays: Union[np.ndarray, Sequence],
    numpy_indexing: bool = True,
    seed: Optional[int] = None,
):
    """Shuffles multiple arrays or lists in sync. Useful for shuffling the data
    and labels in a dataset separately while keeping them synchronised.

    Parameters
    ----------
    arrays, iterable of array-like
        The arrays to shuffle. They must all have the same size of first
        dimension.
    numpy_indexing: bool, default = True
        Whether to use NumPy-style indexing or list comprehension.

    Returns:
    shuffled_arrays: iterable of array-like
        The shuffled arrays.
    """
    if any(len(arrays[0]) != len(x) for x in arrays):
        raise ValueError("Not all arrays have equal first dimension.")

    perm = np.random.default_rng(seed).permutation(len(arrays[0]))
    new_arrays = [
        array[perm] if numpy_indexing else [array[i] for i in perm] for array in arrays
    ]
    return new_arrays


def batch_arrays_by_length(
    arrays_x: Union[np.ndarray, List[np.ndarray]],
    y: np.ndarray,
    batch_size: int = 32,
    shuffle: bool = True,
    uniform_batch_size: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """Batches a list of arrays of different sizes, grouping them by
    size. This is designed for use with variable length sequences. Each
    batch will have a maximum of batch_size arrays, but may have less if
    there are fewer arrays of the same length. It is recommended to use
    the pad_arrays() method of the LabelledDataset instance before using
    this function, in order to quantise the lengths.

    Parameters
    ----------
    arrays_x: list of ndarray
        A list of N-D arrays, possibly of different lengths, to batch.
        The assumption is that all the arrays have the same rank and
        only axis 0 differs in length.
    y: ndarray
        The labels for each of the arrays in arrays_x.
    batch_size: int
        Arrays will be grouped together by size, up to a maximum of
        batch_size, after which a new batch will be created. Thus each
        batch produced will have between 1 and batch_size items.
    shuffle: bool, default = True
        Whether to shuffle array order in a batch.
    uniform_batch_size: bool, default = False
        Whether to keep all batches the same size, batch_size, and pad
        with zeros if necessary, or have batches of different sizes if
        there aren't enough sequences to group together.

    Returns
    -------
    x_list: ndarray,
        The batched arrays. x_list[i] is the i'th
        batch, having between 1 and batch_size items, each of length
        lengths[i].
    y_list: ndarray
        The batched labels corresponding to sequences in x_list.
        y_list[i] has the same length as x_list[i].
    """
    if shuffle:
        arrays_x, y = shuffle_multiple(arrays_x, y, numpy_indexing=False)

    # TODO: implement batch generator

    fixed_shape = arrays_x[0].shape[1:]
    lengths = [x.shape[0] for x in arrays_x]
    unique_len = np.unique(lengths)
    x_dtype = arrays_x[0].dtype
    y_dtype = y.dtype

    xlist = []
    ylist = []
    for length in unique_len:
        idx = np.nonzero(lengths == length)[0]
        for b in range(0, len(idx), batch_size):
            batch_idx = idx[b : b + batch_size]
            size = batch_size if uniform_batch_size else len(batch_idx)
            _x = np.zeros((size, length) + fixed_shape, dtype=x_dtype)
            _y = np.zeros(size, dtype=y_dtype)
            _y[:size] = y[batch_idx]
            for i, j in enumerate(batch_idx):
                _x[i, ...] = arrays_x[j]
            xlist.append(_x)
            ylist.append(_y)
    x_batch = make_array_array(xlist)
    if uniform_batch_size:
        y_batch = np.array(ylist, dtype=y_dtype)
    else:
        y_batch = make_array_array(ylist)
    return x_batch, y_batch


def flat_to_inst(x: np.ndarray, slices: Union[np.ndarray, List[int]]) -> np.ndarray:
    """Takes a concatenated 2D data array and converts it to either a
    contiguous 2D/3D array or a variable-length 3D array, with one
    feature vector/matrix per instance.

    Parameters
    ----------
    x: numpy.ndarray
        Contiguous 2D array containing concatenated instance data.
    slices: list or numpy.ndarray
        Array containing lengths of sequences in x corresponding to
        instance data.

    Returns
    -------
    inst: numpy.ndarray
        Instance array, which is either contiguous 2D, contiguous 3D or
        an array of 2D arrays with varying lengths. It should be the
        case that `inst[i] == x[s : s + slices[i]]`, where `s =
        sum(slices[:i])`.
    """

    if len(x) == len(slices):
        # 2D contiguous array
        return x
    elif all(x == slices[0] for x in slices):
        # 3D contiguous array
        assert len(x) % len(slices) == 0
        return x.reshape(len(slices), len(x) // len(slices), x[0].shape[-1])
    else:
        # 3D variable length array
        start_idx = np.cumsum(slices)[:-1]
        return make_array_array(np.split(x, start_idx, axis=0))


def inst_to_flat(x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """The inverse of flat_to_inst(). Takes an instance 'matrix' and
    converts to a "flattened" 2D matrix.

    Parameters
    ----------
    x: numpy.ndarray
        The instance matrix, which is either a contiguous 2D or 3D
        array, or an array of 1D or 2D arrays.

    Returns
    -------
    flat: numpy.ndarray
        The contiguous 2D matrix containing data for all instances.
    slices: numpy.ndarray
        Array of slices (sequence lengths) in `flat`. It should be the
        case that `x[i] == flat[s : s + slices[i]]`, where `s =
        sum(slices[:i])`.
    """

    slices = np.ones(len(x), dtype=int)
    if len(x.shape) != 2:
        if x[0].ndim == 1:
            # Array of 1D arrays
            x = np.stack(x)  # type: ignore
        else:
            slices = np.array([len(_x) for _x in x])
            if len(x.shape) == 3:
                # Contiguous 3D array
                x = x.reshape(sum(slices), x.shape[2])
            elif x[0].base is not None and all(_x.base is x[0].base for _x in x):
                # Array of views into contiguous array
                x = x[0].base
            else:
                # Array of separate arrays
                x = np.concatenate(x)
    assert sum(slices) == len(x)
    return x, slices



import logging
import time
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    get_scorer,
    make_scorer,
    precision_score,
    recall_score,
)
from sklearn.model_selection import BaseCrossValidator, LeaveOneGroupOut
from sklearn.utils.multiclass import unique_labels

from ertk.dataset import LabelledDataset
from ertk.train import get_cv_splitter, scores_to_df


def binary_accuracy_score(
    y_true, y_pred, *, labels=None, average="binary", sample_weight=None
):
    """Calculated binary accuracy. Binary accuracy is the same as
    accuracy considering only a single class.

    Parameters
    ----------
    y_true:
        Ground truth labels.
    y_pred:
        Predicted labels.
    labels: list, optional
        Labels to include for average != "binary". If None, all unique
        labels in y_true or y_pred are included.
    average: str, optional
        Method to compute average. If "binary" then simply return
        accuracy. If "macro" then return mean binary accuracy. If
        "weighted" then weight the mean binary accuracy by ground truth
        support. If None then return an array of values, one for each
        label in labels.

    Returns
    -------
    label_accs: float or list
        Binary accuracies for labels in labels or average if `average`
        is not None.
    """
    all_labels = unique_labels(y_true, y_pred)
    if average == "binary":
        if len(all_labels) != 2:
            raise ValueError("Must only have two labels when `average` is 'binary'.")
        return accuracy_score(y_true, y_pred, sample_weight=sample_weight)

    if y_true.ndim != 1 or y_pred.ndim != 1:
        raise ValueError("y_true and y_pred must be 1D arrays.")
    if labels is None:
        labels = all_labels
    accs = {l: 0 for l in labels}
    for lab in all_labels:
        acc = accuracy_score(y_true == lab, y_pred == lab, sample_weight=sample_weight)
        accs[lab] = acc
    label_accs = [accs[l] for l in labels]
    if average == "macro":
        return np.mean(label_accs)
    elif average == "weighted":
        counts = [np.count_nonzero(y_true == l) for l in labels]
        return np.average(label_accs, weights=counts)
    elif average is None:
        return label_accs[0] if len(label_accs) == 1 else label_accs


def standard_class_scoring(classes: Sequence[str]):
    """Given a list of classes, returns scikit-learn scorers for overall
    metrics and per-class metrics, for multiclass classification.

    The metrics are:
        Overall:
            uar: Balanced accuracy
            war: Accuracy
            microf1: Micro averaged F1 score
            macrof1: Macro averaged F1 score
        Per-class:
            {class}_rec: Recall
            {class}_prec: Precision
            {class}_f1: F1 score
            {class}_ba: Binary accuracy

    Binary accuracy is calculated as the accuracy when considering
    labels for a class in a one-vs-rest scenario (i.e. the sum of cell
    (i, i) in the confusion matrix and all other cells not in row or
    column i).
    """

    # FIXME: Use labels param with macro and micro recall instead of
    # (balanced) accuracy in order to account for absent classes from
    # test data? Or rely on user to create valid train/test splits.
    scoring = {
        "uar": get_scorer("balanced_accuracy"),
        "war": get_scorer("accuracy"),
        "microf1": make_scorer(f1_score, average="micro"),
        "macrof1": make_scorer(f1_score, average="macro"),
    }
    for i, c in enumerate(classes):
        scoring.update(
            {
                f"{c}_rec": make_scorer(
                    recall_score, average=None, labels=[i], zero_division=0
                ),
                f"{c}_prec": make_scorer(
                    precision_score, average=None, labels=[i], zero_division=0
                ),
                f"{c}_f1": make_scorer(
                    f1_score, average=None, labels=[i], zero_division=0
                ),
                f"{c}_ba": make_scorer(binary_accuracy_score, average=None, labels=[i]),
            }
        )
    return scoring


def dataset_cross_validation(
    clf,
    dataset: LabelledDataset,
    clf_lib: Optional[str] = None,
    partition: Optional[str] = None,
    label: str = "label",
    cv: Union[BaseCrossValidator, int] = 10,
    verbose: int = 0,
    n_jobs: int = 1,
    scoring=None,
    fit_params: Dict[str, Any] = {},
) -> pd.DataFrame:
    """Cross validates a `Classifier` instance on a single dataset.

    Parameters
    ----------
    clf: class that implements fit() and predict()
        The classifier to test.
    dataset: LabelledDataset
        The dataset for within-corpus cross-validation.
    clf_lib: str
        One of {"sk", "tf", "pt"} to select which library-specific
        cross-validation method to use, since they're not all quite
        compatible.
    partition: str, optional
        The name of the partition to cross-validate over. If None, then
        don't use group cross-validation.
    label: str
        The annotations to use as class labels.
    cv: int or BaseCrossValidator
        A splitter used for cross-validation. Default is KFold(10) for
        10 fold cross-validation.
    verbose: bool
        Passed to cross_validate().
    n_jobs: bool
        Passed to cross_validate().
    scoring: str, list, dict, optional
        Scoring metric(s) to use. Can be anything accepted by
        scikit-learn's cross_val* methods (i.e. str, list or dict).
    fit_params: dict
        Additional parameters passed to the model's fit() method. This
        should be used to pass any more specific parameters not covered
        here.

    Returns
    -------
    df: pandas.DataFrame
        A dataframe holding the results from all runs with this model.
    """
    groups = None if partition is None else dataset.get_group_indices(partition)
    if isinstance(cv, int):
        cv = get_cv_splitter(bool(partition), cv)
    if scoring is None:
        scoring = standard_class_scoring(dataset.get_group_names(label))

    cross_validate_fn: Callable[..., Dict[str, Any]]
    if clf_lib == "sk":
        from ertk.sklearn.classification import sk_cross_validate

        # We have to set n_jobs here because using a
        # `with joblib.Parallel...` clause doesn't work properly
        cross_validate_fn = partial(sk_cross_validate, n_jobs=n_jobs)
    elif clf_lib == "tf":
        from ertk.tensorflow.classification import tf_cross_validate

        n_jobs = 1
        cross_validate_fn = tf_cross_validate
    elif clf_lib == "pt":
        from ertk.pytorch.classification import pt_cross_validate

        n_jobs = 1
        cross_validate_fn = pt_cross_validate

    start_time = time.perf_counter()
    logging.info(f"Starting cross-validation with n_jobs={n_jobs}")
    scores = cross_validate_fn(
        clf,
        dataset.x,
        dataset.get_group_indices(label),
        cv=cv,
        scoring=scoring,
        groups=groups,
        verbose=verbose,
        fit_params=fit_params,
    )
    total_time = time.perf_counter() - start_time
    logging.info(f"Cross-validation complete in {total_time:.2f}s")

    index = None
    if isinstance(cv, LeaveOneGroupOut) and partition is not None:
        index = dataset.get_group_names(partition)
    return scores_to_df(scores, index=index)


def train_val_test(
    clf,
    dataset: LabelledDataset,
    train_idx: Union[Sequence[int], np.ndarray],
    valid_idx: Union[Sequence[int], np.ndarray],
    test_idx: Union[Sequence[int], np.ndarray, None] = None,
    label: str = "label",
    clf_lib: Optional[str] = None,
    sample_weight: np.ndarray = None,
    verbose: int = 0,
    scoring=None,
    fit_params: Dict[str, Any] = {},
) -> pd.DataFrame:
    """Trains a `Classifier` instance on some training data, optionally
    using validation data, and returns results on given test data.

    Parameters
    ----------
    clf: class that implements fit() and predict()
        The classifier to test.
    dataset: LabelledDataset
        The dataset for within-corpus cross-validation.
    clf_lib: str
        One of {"sk", "tf", "pt"} to select which library-specific
        cross-validation method to use, since they're not all quite
        compatible.
    verbose: bool
        Passed to train_val_test().
    scoring: str, list, dict, optional
        Scoring metric(s) to use. Can be anything accepted by
        scikit-learn's cross_val* methods (i.e. str, list or dict).
    fit_params: dict
        Additional parameters passed to the model's fit() method. This
        should be used to pass any more specific parameters not covered
        here.

    Returns
    -------
    df: pandas.DataFrame
        A dataframe holding the results from all runs with this model.
    """
    if scoring is None:
        scoring = standard_class_scoring(dataset.get_group_names(label))

    train_val_test_fn: Callable[..., Dict[str, Any]]
    if clf_lib == "sk":
        from ertk.sklearn.classification import sk_train_val_test

        train_val_test_fn = sk_train_val_test
    elif clf_lib == "tf":
        from ertk.tensorflow.classification import tf_train_val_test

        train_val_test_fn = tf_train_val_test
    elif clf_lib == "pt":
        from ertk.pytorch.classification import pt_train_val_test

        train_val_test_fn = pt_train_val_test

    train_idx = np.array(train_idx, copy=False)
    valid_idx = np.array(valid_idx, copy=False)
    if test_idx is None:
        test_idx = valid_idx
    test_idx = np.array(test_idx, copy=False)

    if len(train_idx) == 0 or len(valid_idx) == 0 or len(test_idx) == 0:
        raise ValueError("One of {train, val, test} indices are missing.")

    y = dataset.get_group_indices(label)
    train_data: Tuple = (dataset.x[train_idx], y[train_idx])
    if sample_weight is not None:
        train_data = train_data + (sample_weight[train_idx],)
    valid_data: Tuple = (dataset.x[valid_idx], y[valid_idx])
    if sample_weight is not None:
        valid_data = valid_data + (sample_weight[valid_idx],)
    test_data: Tuple = (dataset.x[test_idx], y[test_idx])
    if sample_weight is not None:
        test_data = test_data + (sample_weight[test_idx],)

    start_time = time.perf_counter()
    logging.info("Starting train/val/test.")
    scores = train_val_test_fn(
        clf,
        train_data,
        valid_data,
        test_data,
        scoring=scoring,
        verbose=verbose,
        fit_params=fit_params,
    )
    total_time = time.perf_counter() - start_time
    logging.info(f"Train/val/test complete in {total_time:.2f}s")

    return scores_to_df(scores)


def get_balanced_sample_weights(labels: Union[List[int], np.ndarray]):
    """Gets sample weights such that each unique label has the same
    total weight across all instances.

    Parameters
    ----------
    labels: list or array
        Sequence of labels of length n_samples.

    Returns
    -------
    sample_weights: np.ndarray
        Array of sample weights of length n_samples.
    """
    unique, indices, counts = np.unique(labels, return_counts=True, return_inverse=True)
    class_weight = len(labels) / (len(unique) * counts)
    sample_weights = class_weight[indices].astype(np.float32)
    return sample_weights



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict

import numpy as np
import torch
import torch.nn.functional as F
from fairseq.checkpoint_utils import load_model_ensemble_and_task
from omegaconf import MISSING

from ertk.config import ERTKConfig

from .base import AudioClipProcessor, FeatureExtractor


class _Aggregation(Enum):
    MEAN = "mean"
    MAX = "max"
    NONE = "none"


@dataclass
class FairseqExtractorConfig(ERTKConfig):
    model_type: str = MISSING
    checkpoint: str = MISSING
    layer: str = "context"
    aggregate: _Aggregation = _Aggregation.MEAN
    device: str = "cuda"
    arg_overrides: Dict[str, Any] = field(default_factory=dict)


class FairseqExtractor(
    FeatureExtractor, AudioClipProcessor, fname="fairseq", config=FairseqExtractorConfig
):
    config: FairseqExtractorConfig

    def __init__(self, config: FairseqExtractorConfig) -> None:
        super().__init__(config)

        print(f"Loading model from {config.checkpoint}")
        [model], args, task = load_model_ensemble_and_task(
            [config.checkpoint], arg_overrides=config.arg_overrides
        )
        model.to(device=config.device)
        model.eval()
        self.model = model
        self.args = args
        self.task = task
        torch.set_grad_enabled(False)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if kwargs.pop("sr") != 16000:
            raise ValueError("sample rate should be 16kHz")

        tensor = torch.tensor(x, device=self.config.device).unsqueeze(0)
        if hasattr(self.task, "normalize") and self.task.normalize:
            tensor = F.layer_norm(tensor, tensor.shape)
        if self.config.model_type == "wav2vec":
            z = self.model.feature_extractor(tensor)
            if self.model.vector_quantizer is not None:
                z, _ = self.model.vector_quantizer.forward_idx(z)
            feats = z
            if self.config.layer == "context":
                feats = self.model.feature_aggregator(feats)
        elif self.config.model_type == "wav2vec2":
            if self.config.layer == "context":
                c = self.model.extract_features(tensor, None)["x"]
                # Transpose to (batch, feats, steps)
                feats = c.transpose(-2, -1)
            else:
                feats = self.model.feature_extractor(tensor)
        else:  # HuBERT
            if self.config.layer == "context":
                c, _ = self.model.extract_features(tensor)
                # Transpose to (batch, feats, steps)
                feats = c.transpose(-2, -1)
            else:
                feats = self.model.feature_extractor(tensor)

        if self.config.aggregate == _Aggregation.NONE:
            return feats[0].transpose(1, 0).cpu().numpy()
        elif self.config.aggregate == _Aggregation.MEAN:
            return feats[0].mean(-1).cpu().numpy()
        elif self.config.aggregate == _Aggregation.MAX:
            return feats[0].max(-1).cpu().numpy()
        else:
            raise ValueError(f"aggregation type {self.config.aggregate} not supported")

    @property
    def dim(self) -> int:
        return 1024

    @property
    def is_sequence(self) -> bool:
        return self.config.aggregate == _Aggregation.NONE



from typing import Any, Dict

import torch
import torch.nn.functional as F
from fairseq.checkpoint_utils import load_model_ensemble_and_task
from torch import nn
from torch.optim import SGD, Adam, RMSprop

from ._base import SimpleClassificationModel


class Model(SimpleClassificationModel):
    def __init__(
        self,
        ckpt: str,
        n_classes: int,
        n_features: int = 1,
        opt: str = "sgd",
        lr: float = 0.1,
        opt_params: Dict[str, Any] = {},
    ) -> None:
        super().__init__(n_classes=n_classes, n_features=1)

        [model], args, task = load_model_ensemble_and_task([ckpt])
        self.wav2vec = model
        self.w2v_args = args
        # self.w2v_task = task

        # Needed for multi-GPU training because of
        # https://github.com/pytorch/fairseq/issues/3482
        self.w2v_cfg = dict(task.cfg)
        if "text_compression_level" in self.w2v_cfg:
            self.w2v_cfg["text_compression_level"] = str(
                self.w2v_cfg["text_compression_level"]
            )

        self.lr = lr
        self.opt = opt
        self.opt_params = dict(opt_params)
        self.save_hyperparameters("ckpt", "lr", "opt", "opt_params")

        self.final = nn.Linear(512, n_classes)

    def forward(self, x):
        x = x.squeeze(-1)
        x = self.wav2vec.feature_extractor(x)
        if self.wav2vec.vector_quantizer is not None:
            x, _ = self.wav2vec.vector_quantizer.forward_idx(x)
        x = x.mean(-1)
        x = self.final(x)
        return x

    def configure_optimizers(self):
        opt_cls = {"adam": Adam, "rmsprop": RMSprop, "sgd": SGD}[self.opt]
        return opt_cls(self.parameters(), lr=self.lr, **self.opt_params)

    def preprocess_input(self, x: torch.Tensor) -> torch.Tensor:
        if self.w2v_cfg["normalize"]:
            with torch.no_grad():
                x = F.layer_norm(x, x.shape)
        return x



import re
import warnings
from pathlib import Path
from typing import Tuple

import click
import librosa
import soundfile
from joblib import delayed

from ertk.utils import TqdmParallel

REGEX = re.compile(
    r"^\*([A-Z]{3}):\t.*(?:\n\t.*)*[.?!](?: .*\x15(\d+)_(\d+)\x15)?$", re.MULTILINE
)


@click.command()
@click.argument("input", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.argument("output", type=Path)
@click.option("--prefix", type=str, default="")
def main(input: Tuple[Path], output: Path, prefix: str):
    """Splits CHAT (.cha) files and associated audio into segments."""

    def process(path: Path):
        warnings.simplefilter("ignore", UserWarning)
        audio, _ = librosa.load(path, sr=16000, res_type="kaiser_fast")
        warnings.simplefilter("default", UserWarning)

        cha_file = path.with_suffix(".cha")
        with open(cha_file) as fid:
            for i, match in enumerate(REGEX.finditer(fid.read())):
                if match[2] is None or match[3] is None:
                    # This utterance doesn't have timecode info
                    continue

                s_sam = int(match[2]) * 16
                e_sam = int(match[3]) * 16
                if s_sam > len(audio):
                    print(f"WARNING: audio {path} shorter than expected.")
                    break
                out_name = f"{prefix}{path.stem}_{i + 1:03d}_{match[1]}"
                split = audio[s_sam:e_sam]
                out_file = output / (out_name + ".wav")
                soundfile.write(out_file, split, 16000)

    output.mkdir(parents=True, exist_ok=True)
    all_files = [p for path in input for p in path.glob("**/*.mp3")]
    TqdmParallel(total=len(all_files), desc="Processing CHAT files", n_jobs=-1)(
        map(delayed(process), all_files)
    )


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F
from torch import nn

from .layers import Attention1D


def frame(
    x: torch.Tensor, frame_width: int, frame_step: int, pad: bool = False
) -> torch.Tensor:
    """Create a contiguous tensor of frames of x, where axis 1 of x is
    considered the time axis.

    Parameters
    ----------
    x: torch.Tensor
        A tensor of shape (B, T, ...) where B is the batch size, T is
        the length of the sequence.
    frame_width: int
        The width of each frame in steps.
    frame_step: int
        The number of steps each frame is shifted along the time axis.
    pad: bool
        Whether to pad the time axis or not, if T cannot evenly fit all
        frames.

    Returns
    -------
    frames: torch.Tensor
        A tensor of shape (B, N, frame_width, ...) where N is the number
        of frames. It will have one more dimension than x, and the final
        D - 3 dimensions will be the same as for x.
        N = 1 + int(pad) + (T - frame_width) / frame_step
    """
    batch_size, steps, *_ = x.size()
    if steps < frame_width:
        raise ValueError("Length of input sequence too small.")
    if (steps % frame_step) != (frame_width % frame_step):
        if pad:
            right_pad = frame_step - (steps % frame_step)
            x = F.pad(x, (0, 0) * (x.ndim - 2) + (0, right_pad))
        else:
            x = x[:, steps - (steps % frame_step), ...]
    steps -= steps % frame_step

    n_frames = 1 + int(pad) + (steps - frame_width) // frame_step
    new_shape = torch.Size((batch_size, n_frames, frame_width) + x.size()[2:])
    frames = torch.empty(new_shape, device=x.device, requires_grad=False)
    for i in range(n_frames):
        start = i * frame_step
        frames[:, i, :, ...] = x[:, start : start + frame_width, ...]
    return frames


class Model(nn.Module):
    def __init__(self, n_classes: int, n_features: int = 1) -> None:
        super().__init__()

        self.dropout = nn.Dropout(0.1)
        self.conv1 = nn.Sequential(
            nn.Conv1d(1, 40, kernel_size=40, padding=20), nn.ReLU(inplace=True)
        )
        self.maxpool1 = nn.MaxPool1d(2)
        self.conv2 = nn.Sequential(
            nn.Conv1d(40, 40, kernel_size=40, padding=20), nn.ReLU(inplace=True)
        )
        self.maxpool2 = nn.MaxPool1d(10)

        self.gru = nn.GRU(1280, 128, num_layers=2, batch_first=True)
        self.attention = Attention1D(128)
        self.predict = nn.Linear(128, n_classes)

    def forward(self, x: torch.Tensor):
        # x has shape (batch, steps, 640)
        batch_size, steps, _ = x.size()
        x = self.dropout(x).view(batch_size * steps, 1, 640)

        # Padding is even so we need to cut the last value
        extraction = self.conv1(x)[:, :, :-1]
        extraction = self.maxpool1(extraction)
        extraction = self.conv2(extraction)[:, :, :-1]
        # (batch * steps, 40, 320) -> (batch * steps, 320, 40)
        extraction = extraction.transpose(1, 2)
        extraction = self.maxpool2(extraction)
        extraction = extraction.transpose(1, 2).view(batch_size, steps, 1280)

        h0 = torch.zeros(2, batch_size, 128, dtype=x.dtype, device=x.device)
        gru, _ = self.gru(extraction, h0)
        att = self.attention(gru)
        out = self.predict(att)
        return out

    def preprocess_input(self, x: torch.Tensor):
        return frame(x, 640, 160).squeeze(-1)



"""Basic MLP model."""

from typing import Sequence

from tensorflow.keras.layers import Dense, Dropout, Input
from tensorflow.keras.models import Model


def model(
    n_features: int,
    n_classes: int,
    units: Sequence[int] = [512],
    dropout: float = 0.5,
) -> Model:
    """Creates a Keras model with hidden layers and ReLU activation,
    with 50% dropout.
    """
    inputs = Input((n_features,), name="input")
    x = inputs
    for n_units in units:
        x = Dense(n_units, activation="relu")(x)
        x = Dropout(dropout)(x)
    x = Dense(n_classes, activation="softmax")(x)
    return Model(inputs=inputs, outputs=x)



import math

import torch
import torch.nn.functional as F
from torch import nn

nn.init.normal_


class Attention1D(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super().__init__()

        self.weight = nn.Parameter(torch.randn(in_channels, 1) / math.sqrt(in_channels))

    def forward(self, x: torch.Tensor):
        # x : (..., length, in_channels)
        alpha = F.softmax(x.matmul(self.weight), -2)
        # alpha : (..., length, 1)
        r = torch.sum(alpha * x, -2)
        return r



# Copyright 2019 The TensorFlow Authors All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Hyperparameters for YAMNet."""

from dataclasses import dataclass


# The following hyperparameters (except patch_hop_seconds) were used to train YAMNet,
# so expect some variability in performance if you change these. The patch hop can
# be changed arbitrarily: a smaller hop should give you more patches from the same
# clip and possibly better performance at a larger computational cost.
@dataclass(frozen=True)  # Instances of this class are immutable.
class Params:
    sample_rate: float = 16000.0
    stft_window_seconds: float = 0.025
    stft_hop_seconds: float = 0.010
    mel_bands: int = 64
    mel_min_hz: float = 125.0
    mel_max_hz: float = 7500.0
    log_offset: float = 0.001
    patch_window_seconds: float = 0.96
    patch_hop_seconds: float = 0.48

    @property
    def patch_frames(self):
        return int(round(self.patch_window_seconds / self.stft_hop_seconds))

    @property
    def patch_bands(self):
        return self.mel_bands

    num_classes: int = 521
    conv_padding: str = "same"
    batchnorm_center: bool = True
    batchnorm_scale: bool = False
    batchnorm_epsilon: float = 1e-4
    classifier_activation: str = "sigmoid"

    tflite_compatible: bool = False



import logging
import pprint
from dataclasses import asdict
from io import StringIO
from pathlib import Path
from typing import Tuple

import click
from joblib import delayed
from omegaconf import OmegaConf
from tqdm import tqdm

from ertk.dataset import get_audio_paths, read_features, write_features
from ertk.utils import TqdmParallel


def list_processors(ctx, param, val):
    from ertk.preprocessing import InstanceProcessor

    if not val or ctx.resilient_parsing:
        return

    print(f"\nAvailable processors:\n{InstanceProcessor.valid_preprocessors()}\n")
    print(
        "Below are the default configurations for registered feature extractors. Items "
        "with '???' are required to be specified manually.\n"
    )
    for key in InstanceProcessor.valid_preprocessors():
        print(f"[{key}]: ", end="")
        proc_cls = InstanceProcessor.get_processor_class(key)
        print(", ".join([b.__name__ for b in proc_cls.__bases__]))
        config_cls = proc_cls.get_config_type()
        for k, v in asdict(config_cls()).items():
            print(f"  {k} = {v}")
        print()

    ctx.exit()


@click.command()
@click.argument("input", type=click.Path(exists=True, path_type=Path))
@click.argument("output", type=Path)
@click.option("--features", required=True, help="Features to extract.")
@click.option(
    "--list_processors",
    is_flag=True,
    callback=list_processors,
    expose_value=False,
    is_eager=True,
    help="Show options for registered feature extractors.",
)
@click.option(
    "--config",
    type=click.Path(exists=True, path_type=Path),
    help="Extractor config file.",
)
@click.option("--corpus", default="", help="Corpus name to set.")
@click.option(
    "--batch_size",
    type=int,
    default=1,
    help="Batch size for processing. If batch_size is greater than 1, clips will be "
    "batched together.",
)
@click.option(
    "--sample_rate", type=float, help="Resample to this rate for audio input."
)
@click.option("--n_jobs", type=int, default=-1, help="Number of parallel jobs to run.")
@click.option("--verbose", is_flag=True)
@click.argument("restargs", nargs=-1)
def main(
    input: Path,
    output: Path,
    features: str,
    config: Path,
    corpus: str,
    batch_size: int,
    sample_rate: float,
    n_jobs: int,
    verbose: bool,
    restargs: Tuple[str],
):
    """Process features or audio files in INPUT, write features to
    OUTPUT.
    """

    from ertk.preprocessing import InstanceProcessor

    if verbose:
        logging.basicConfig(level=logging.INFO)

    extractor_cls = InstanceProcessor.get_processor_class(features)
    config_cls = extractor_cls.get_config_type()
    if config:
        conf = OmegaConf.load(config)
    else:
        conf = extractor_cls.get_default_config()
    conf = OmegaConf.merge(conf, OmegaConf.from_dotlist(list(restargs)))
    if verbose:
        print("Using configuration:")
        pprint.pprint(dict(conf))
    extractor = extractor_cls(config_cls.from_config(conf))

    if input.suffix == ".txt":
        # TODO: unify with read_features()
        paths = get_audio_paths(input)
        names = [Path(x).stem for x in paths]
        if batch_size != 1:
            feats = list(
                tqdm(
                    extractor.process_files(
                        paths, batch_size=batch_size, sr=sample_rate
                    ),
                    total=len(paths),
                    desc="Processing files",
                )
            )
        else:
            feats = TqdmParallel(len(paths), "Processing files", n_jobs=n_jobs)(
                delayed(extractor.process_file)(x, sr=sample_rate) for x in paths
            )
    else:
        input_data = read_features(input)
        corpus = corpus or input_data.corpus
        names = input_data.names
        if batch_size != 1:
            feats = list(
                tqdm(
                    extractor.process_all(input_data.features, batch_size),
                    total=len(input_data),
                    desc="Processing files",
                )
            )
        else:
            feats = TqdmParallel(len(input_data), "Processing files", n_jobs=n_jobs)(
                map(delayed(extractor.process_instance), input_data.features)
            )
    write_features(output, feats, names=names, corpus=corpus)
    print(f"Wrote features to {output}")



from typing import Any, Callable, Dict, Iterable, List, Union

import numpy as np
import pandas as pd
from sklearn.base import MetaEstimatorMixin
from sklearn.metrics import get_scorer
from sklearn.model_selection import (
    BaseCrossValidator,
    GroupKFold,
    GroupShuffleSplit,
    LeaveOneGroupOut,
    LeaveOneOut,
    StratifiedKFold,
    StratifiedShuffleSplit,
)
from sklearn.model_selection._split import _BaseKFold
from sklearn.model_selection._validation import _score
from sklearn.pipeline import Pipeline
from sklearn.utils import check_array

from ertk.utils import ScoreFunction, filter_kwargs


class TrainValidation(BaseCrossValidator):
    """Validation method that uses the training set as validation set."""

    def split(self, X, y, groups):
        yield np.arange(len(X)), np.arange(len(X))

    def get_n_splits(self, X, y, groups):
        return 1


class ShuffleGroupKFold(_BaseKFold):
    """Like GroupKFold but with random combinations of groups instead of
    deterministic combinations based on group size. This is most useful
    if you have groups of near equal size, and you want group k-fold CV,
    where k divides n_groups.

    Note: If shuffle=False, this does not behave identical to
    GroupKFold, but rather splits groups in sorted order (as returned by
    `numpy.unique()`).
    """

    def __init__(self, n_splits=5, *, shuffle=False, random_state=None):
        super().__init__(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

    def _iter_test_indices(self, X, y, groups):
        if groups is None:
            raise ValueError("The 'groups' parameter should not be None.")
        groups = check_array(groups, ensure_2d=False, dtype=None)

        unique_groups, groups = np.unique(groups, return_inverse=True)
        n_groups = len(unique_groups)

        if self.n_splits > n_groups:
            raise ValueError(
                "Cannot have number of splits n_splits=%d greater"
                " than the number of groups: %d." % (self.n_splits, n_groups)
            )

        # Pairs of start,end indices of groups each of n folds
        fold_idx = np.linspace(0, n_groups, self.n_splits + 1, dtype=int)

        group_order = np.arange(n_groups)
        if self.shuffle:
            # Shuffle order groups appear in folds
            group_order = np.random.default_rng(self.random_state).permutation(
                group_order
            )

        # Mapping from group index to fold index
        group_to_fold = np.zeros(n_groups)

        for fold, (g1, g2) in enumerate(zip(fold_idx[:-1], fold_idx[1:])):
            group_to_fold[group_order[g1:g2]] = fold

        indices = group_to_fold[groups]

        for f in range(self.n_splits):
            yield np.where(indices == f)[0]


class ValidationSplit(BaseCrossValidator):
    """Validation method that uses a pre-defined validation set."""

    def __init__(self, valid_idx: Union[List[int], np.ndarray]):
        self.valid_idx = valid_idx

    def split(self, X, y, groups):
        train_idx = np.arange(len(X))
        train_idx = train_idx[~np.isin(train_idx, self.valid_idx)]
        yield train_idx, self.valid_idx

    def get_n_splits(self, X, y, groups):
        return 1


def get_cv_splitter(
    group: bool,
    k: int,
    test_size: float = 0.2,
    shuffle: bool = False,
    random_state: int = None,
) -> BaseCrossValidator:
    """Gets an appropriate cross-validation splitter for the given
    number of folds and groups, or a single random split.

    Parameters
    ----------
    group: bool
        Whether to split over pre-defined groups of instances.
    k: int
        If k > 1 then do k-fold CV. If k == 1 then do one random
        split. If k = -1 then do leave-one-out. If k == 0 then use the
        whole train set as validation split.
    test_size: float
        The size of the test set when k == 1 (one random split).
    shuffle: bool
        Whether to shuffle when using k-fold for k > 1.
    random_state: int, optional
        The random state to set for splitters with shuffling behaviour.

    Returns
    -------
    splitter: BaseCrossValidator
        Cross-validation splitter that has `split()` and
        `get_n_splits()` methods.
    """
    # TODO: Leave-|k|-out for k < 0?
    if k == 0:
        return TrainValidation()
    if group:
        if k > 1:
            if shuffle:
                return ShuffleGroupKFold(k, shuffle=shuffle, random_state=random_state)
            return GroupKFold(k)
        elif k == 1:
            return GroupShuffleSplit(1, test_size=test_size, random_state=random_state)
        return LeaveOneGroupOut()
    if k > 1:
        return StratifiedKFold(k, shuffle=shuffle, random_state=random_state)
    elif k == 1:
        return StratifiedShuffleSplit(1, test_size=test_size, random_state=random_state)
    return LeaveOneOut()


def get_pipeline_params(params: Dict[str, Any], pipeline: Pipeline):
    """Modifies parameter names to pass to a Pipeline instance's `fit()`
    method.

    Parameters
    ----------
    params: dict
        Parameters to pass to Pipeline.fit(). All parameters are passed
        to all estimators in the pipeline so long as they are valid.
    pipeline: Pipeline
        The pipeline instance.

    Returns
    -------
    new_params: dict
        Parameters filtered and prepended with pipeline step names and
        double underscore (e.g. groups -> clf__groups).
    """
    new_params = {}
    for name, est in pipeline.named_steps.items():
        if est is None or est == "passthrough":
            continue
        filt_params = params
        if not isinstance(est, MetaEstimatorMixin):  # May pass through kwargs
            filt_params = filter_kwargs(filt_params, est.fit)
        new_params.update({name + "__" + k: v for k, v in filt_params.items()})
    return new_params


def get_scores(
    scoring: Union[str, List[str], Dict[str, ScoreFunction], Callable[..., float]],
    y_pred: np.ndarray,
    y_true: np.ndarray,
) -> Dict[str, Any]:
    """Get dictionary of scores for predictions.

    Parameters
    ----------
    scoring: str, list, dict or callable
        Score(s) to calculate. This takes the same for as for
        scikit-learn's cross_val_* methods.
    y_pred: array-like
        Predictions.
    y_true: array-like
        Ground truth.

    Returns
    -------
    scores: dict
        A dictionary mapping score names to corresponding score(s).
    """

    class DummyEstimator:
        """Class that implements a dummy estimator for scoring, to avoid
        repeated invocations of `predict()` etc.
        """

        def __init__(self, y_pred):
            self.y_pred = y_pred

        def predict(self, x, **kwargs):
            return self.y_pred

        def predict_proba(self, x, **kwargs):
            return self.y_pred

        def decision_function(self, x, **kwargs):
            return self.y_pred

    y_pred = np.array(y_pred)
    y_true = np.array(y_true)
    dummy = DummyEstimator(y_pred)

    if isinstance(scoring, str):
        scoring = {"score": get_scorer(scoring)}
    elif callable(scoring):
        scoring = {"score": scoring}
    elif isinstance(scoring, list):
        scoring = {x: get_scorer(x) for x in scoring}
    return _score(dummy, None, y_true, scoring)


def scores_to_df(
    scores: Dict[str, np.ndarray], index: Union[pd.Index, Iterable, None] = None
):
    """Convert scikit-learn scores dictionary to pandas dataframe.

    Parameters
    ----------
    scores: dict
        Scores dictionary. Each key is a string and all values are
        either scalar or arrays.
    index: pd.Index or iterable, optional
        Optional index to use. Must have the same length as each value
        in `scores`. If not given a `pd.RangeIndex` is used.

    Returns
    -------
    pd.DataFrame
        Scores dataframe, with a column per key in `scores` and row per
        item in each value array.
    """
    _val = next(iter(scores.values()))
    if index is None:
        if np.ndim(_val) == 0:
            index = pd.RangeIndex(1, name="fold")
        else:
            index = pd.RangeIndex(len(_val), name="fold")
    elif not isinstance(index, pd.Index):
        index = pd.Index(index)
    return pd.DataFrame(
        {k[5:] if k.startswith("test_") else k: v for k, v in scores.items()},
        index=index,
    )



import tensorflow as tf
from tensorflow.keras.layers import Layer


class Attention1D(Layer):
    """Layer that implements simple weighted pooling using softmax
    attention over a sequence of input vectors.
    """

    def build(self, input_shape: tuple):
        _, _, size = input_shape
        self.weight = self.add_weight("weight", (size, 1))

    def call(self, inputs, **kwargs):
        alpha = tf.matmul(inputs, self.weight)  # (batch, steps, 1)
        alpha = tf.nn.softmax(alpha, axis=-2)
        r = tf.reduce_sum(alpha * inputs, axis=-2)  # (batch, size)
        return r



import click

from .class_cv import main as class_cv_main
from .class_tvt import main as tvt_main
from .classify import main as classify_main


@click.group(invoke_without_command=True)
def cli():
    """ERTK command-line scripts."""
    click.echo("ERTK CLI")


cli.add_command(classify_main, "classify")
cli.add_command(class_cv_main, "class_cv")
cli.add_command(tvt_main, "class_tvt")



from pathlib import Path

import click
import pandas as pd

from ertk.dataset import get_audio_paths


@click.command()
@click.argument("clips", type=click.Path(exists=True, path_type=Path))
@click.argument("filelist", type=click.Path(exists=True, path_type=Path))
def main(clips: Path, filelist: Path):
    """Convert names in CLIPS to filepaths from FILELIST. Write to
    stdout.
    """

    if clips.suffix == ".csv":
        df = pd.read_csv(clips, header=0, index_col=0)
        names = list(df.index)
    else:
        with open(clips) as fid:
            names = list(map(str.strip, fid))

    paths = get_audio_paths(filelist)
    name_to_file = {p.stem: str(p) for p in paths}

    print("\n".join(map(name_to_file.__getitem__, names)))


if __name__ == "__main__":
    main()
